/*
 * gui.c
 * Xlib display routines.
 * Milos Glisic, 2000.
 */

#include <stdio.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include "config.h"
#include "proto.h"
#include "splash.h"


static Display     *display;
static int          screen;
static int          depth;
static Window       window;
static GC           gc;
static XImage      *real, *image = NULL, *splash;
static Colormap     cmap;
static Cursor       cursor;



/* An error handler that ignores errors. For internal use. */
int gui_error_handler(Display *disp, XErrorEvent *err)
{
  /* fprintf(stderr, "Error: %d\n", err->error_code); */
  return 0;
}


/* Event loop. */
void gui_event_loop()
{
  XEvent event;
  char xlat[20];
  int count;
  KeySym key;

  while(1) {

    fprintf(stderr, "waiting...");

    if(XPending(display) > 0) {
      XNextEvent(display, &event);
      fprintf(stderr, "an event!\n");
    }
    else {
      fprintf(stderr, "no event!\n");
      continue;
    }

    if(event.type == KeyPress) {
      fprintf(stderr, "keypress!\n");

      count = XLookupString((XKeyEvent *)&event, xlat, 20, &key, NULL);

      if(key == XK_Q || key == XK_q) {
	gui_kill();
      }
    }
  }
}


/* Initialize the display window. Returns whether it worked. */
int gui_init()
{
  XSetWindowAttributes  winattr;
  XWindowAttributes     rootwin;
  unsigned long         winmask;
  XSizeHints           *sh;
  XClassHint           *ch;
  XGCValues             gcvalues;
  int                   x, y;
  unsigned int          screenw, width = WIDTH;
  unsigned int          screenh, height = HEIGHT;
  int                   border_width = 1;
  XWMHints             *xwmh;
  XTextProperty         wname, iname;
  char                 *window_title = NAME;


  display = XOpenDisplay(NULL);
  if(display == NULL) {
    XDisplayName(NULL);
    return 0;
  }

  screen = DefaultScreen(display);
  depth = DefaultDepth(display, screen);

  /* Get info on the root window. */
  XGetWindowAttributes(display, DefaultRootWindow(display), &rootwin);
  screenw = rootwin.width;
  screenh = rootwin.height;

  /* Position our window at the centre of the screen. */
  x = screenw / 2 - WIDTH / 2;
  y = screenh / 2 - HEIGHT / 2;

  /* Set some window attributes. */
  winattr.border_pixel = BlackPixel(display, screen);
  winattr.background_pixel = BlackPixel(display, screen);
  winattr.event_mask = KeyPress;
  winattr.do_not_propagate_mask = ButtonPress | ButtonRelease;
  winattr.override_redirect = False;
  winmask = CWBackPixel | CWBorderPixel | CWOverrideRedirect |
    CWDontPropagate | CWEventMask;

  /* Create the window. */
  window = XCreateWindow(display, RootWindow(display, screen),
                    x, y, width, height, border_width,
                    depth, InputOutput, CopyFromParent,
                    winmask, &winattr);

  if(XStringListToTextProperty(&window_title, 1, &wname)==0 ||
     XStringListToTextProperty(&window_title, 1, &iname)==0 ||
     (xwmh = XAllocWMHints()) == NULL ||
     (sh = XAllocSizeHints()) == NULL ||
     (ch = XAllocClassHint()) == NULL) {

    gui_kill();
    return 0;
  }

  /* WM hints... */
  xwmh->flags = (InputHint|StateHint);
  xwmh->input = False;
  xwmh->initial_state = NormalState;

  /* Size hints... */
  sh->flags   = PPosition | PSize;
  sh->x       = x; 
  sh->y       = y; 
  sh->width   = width; 
  sh->height  = height; 
  sh->min_width   = width; 
  sh->min_height  = height; 
  sh->max_width   = width; 
  sh->max_height  = height; 

  XSetNormalHints(display, window, sh);

  /* Class hints... */
  ch->res_name = NAME;
  ch->res_class = NAME;

  XSetWMProperties(display, window, &wname, &iname, NULL, 0, sh, xwmh, ch);

  cmap = DefaultColormap(display, screen);

  /* Set a target-looking cursor for our window. */
  cursor = XCreateFontCursor(display, XC_tcross);
  XDefineCursor(display, window, cursor);

  if((gc = XCreateGC(display, window, (unsigned long) 0, &gcvalues))==0) {
    gui_kill();
    return 0;   
  }

  /* Put the window on the screen. */
  XMapWindow(display, window);
  XSync(display, False);

  /* Ignore X errors at this point. */
  XSetErrorHandler(gui_error_handler);

  /* Keep retrying to capture the image until it works. */
  while((image = XGetImage(display, window, 0, 0,
			   width, height, 0, ZPixmap)) == NULL)
    usleep(20);

#ifdef VERBOSE
  fprintf(stderr, "Initialized display: depth=%dbpp, dim=%dx%d\n",
	  depth, screenw, screenh);
#endif

  return 1;
}


/* Splash screen routine. */
void gui_splash_on()
{
  XpmAttributes xpmattr;

  xpmattr.depth = depth;
  xpmattr.valuemask = XpmDepth;

  XpmCreateImageFromData(display, splash_xpm, &splash, NULL, &xpmattr);

  real = splash;

  gui_refresh();
}


/* Enable game mode (turn off splash) */
void gui_splash_off()
{
  real = image;

  /* Free the splash image. */
  XDestroyImage(splash);
}


/* Draw the background on our image. */
void gui_update_back(int xc, int yc, int **back)
{
  int x, y;

  for(x = 0; x < WIDTH; x++)
    for(y = 0; y < HEIGHT; y++)
      XPutPixel(image, x, y, back[xc + x][yc + y]);
}


/* Set the pixel (cell) at the specified location to a colour. */
void gui_put_pixel(int x, int y, int color)
{
  XPutPixel(image, x, y, color);
}


/* Get the coordinates of the mouse pointer in the window.
   Sets -1, -1 if the cursor is not on the screen.
   Returns whether the right mouse button is clicked. */
int gui_get_mouse(int *x, int *y)
{
  Window root, child;
  int root_x, root_y;
  unsigned int keys_buttons;
  
  XQueryPointer(display, window, &root, &child, &root_x, &root_y,
		x, y, &keys_buttons);
  
  if(keys_buttons & Button3Mask)
    return 2;

  if(keys_buttons & Button1Mask)
    return 1;
  else 
    return 0;
}


/* Get the pixel value for the given RGB combo. */
int gui_rgb(int r, int g, int b)
{
  XColor color;

  color.red = r;
  color.green = g;
  color.blue = b;

  if(!XAllocColor(display, cmap, &color))
    return 0;
  else
    return color.pixel;
}




/* Set every pixel in the XImage to 0. */
void gui_clear_screen()
{
  int a, b;

  for(a = 0; a < WIDTH; a++)
    for(b = 0; b < HEIGHT; b++)
      XPutPixel(image, a, b, 0);
}


/* Synchronize the window with the stored XImage. */
void gui_refresh()
{
  XPutImage(display, window, gc, real, 0, 0, 0, 0, WIDTH, HEIGHT);
}


/* Destroy the window and close the X connection. */
void gui_kill()
{
  if(image == NULL)
    return;

  XDestroyImage(image);
  XFreeGC(display, gc);
  XDestroyWindow(display, window);
  XCloseDisplay(display);
}

