/*
 * main.c
 * Main stuff.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>

#include "config.h"
#include "proto.h"


int fps;


/* Exit and stuff. */
void clean_exit()
{
  net_close();			/* Close the connection */
  gui_kill();			/* Close the gui */

  _exit(0);
}


/* Segmentation fault. */
void segvhandle(int sig)
{
  fprintf(stderr, SEGMSG);
  clean_exit(0);
}


/* Handle a keyboard interrupt. */
void inthandle(int sig)
{
  /* revert to original handler. */
  signal(SIGINT, SIG_DFL);

  /* Stop the ticking. */
  ualarm(0, 0);

  fprintf(stderr, "Terminating...\n");
  clean_exit();
}


/* If we get SIGALRM, start a new game. */
void alarmhandle(int sig)
{
  /* Make the game go. */
  update();

  /* Sync with pal. */
  update_pal();
    
  /* Read up news on the net. */
  net_snag();
}


/* alarm handler for the robot. */
void robalarmhandle(int sig)
{
  /* Do crazy fake-intelligence stuff. */
  rob_update();

  update_pal();

  net_snag();
}


/* Figure out how long it takes this computer to do a game loop. */
int calibrate(int times)
{
  int i, sec, ms, diff = 0;
  struct timeval tv;
  
  gettimeofday(&tv, NULL);

  for(i = 0; i < times; i++) {
    
    sec = tv.tv_sec;
    ms = tv.tv_usec;
    
    update();
    
    gettimeofday(&tv, NULL);
    sec = tv.tv_sec - sec;
    ms = tv.tv_usec - ms;
    
    diff += sec * 1000000 + ms;
  }
  
  return diff / times;
}


/* Calibrate FPS rate and sync with partner. */
int negotiate(int net)
{
  int us, peer, result;
  float peerver;
  char buf[BUFLEN], *arg1, *arg2;


  us = calibrate(CALIB_TIMES);
  
  if(net) {
    sprintf(buf, "I:%s:%d:\n", VERSION, us);
    
    net_put(buf);
    net_get(buf);

    strtok(buf, SEP);
    arg1 = strtok(NULL, SEP);
    arg2 = strtok(NULL, SEP);
    
    if(buf[0] != 'I' || arg1 == NULL || arg2 == NULL)
      return -1;    

#ifdef VERBOSE
    fprintf(stderr, "Peer is running version \"%s\".\n", arg1);
#endif

    /* Check version compatibility. */
    if((peerver = atof(arg1)) < atof(COMPATVER))
      return -2;

    peer = atoi(arg2);
    
    if(peer < us)
      result = 1000000 / us;
    else
      result = 1000000 / peer;
  } else
    result = 1000000 / us;

  /* We'll make sure the framerate doesn't exceed a reasonable ceiling. */
  if(result > MAX_FPS)
    result = MAX_FPS;

  return result;
}


/* Main. */
int main(int argc, char **argv)
{
  int rob = 0, net = 1, pid, difficulty;

  /* Close stdin. */
  close(0);

  fprintf(stderr, "%s %s by %s\n", NAME, VERSION, AUTHOR);

  if(argc < 2) {
    fprintf(stderr, USAGE);
    return 1;
  }
  
  /* Install catastrophe handles. */
  (void) signal(SIGSEGV, segvhandle);
  (void) signal(SIGFPE, segvhandle);
  (void) signal(SIGINT, inthandle);
  (void) signal(SIGPIPE, inthandle);
  
  /* Initialize network stuff. */
  if(strcmp(argv[1], "-s")==0) {
    if(!net_init()) {
      perror("net_init()");
      return 1;
    }
    if(!net_accept()) {
      perror("net_accept()");
      return 1;
    }
  } else if(strcmp(argv[1], "-t")==0) {
    fprintf(stderr, "Playing in test (single player) mode.\n");
    net = 0;
  } else if(strncmp(argv[1], "-r", 2)==0) {

    if(strlen(argv[1]) < 3) {
      fprintf(stderr, USAGE);
      return 1;
    }

    switch(argv[1][2]) {
    case 'e':
      difficulty = 0;
      break;
    case 'm':
      difficulty = 1;
      break;
    case 'h':
      difficulty = 2;
      break;
    default:
      fprintf(stderr, "That difficulty level doesn't exist.\n");
      return 1;
    }

    rob = 1;
    fprintf(stderr, "Starting robot player (Zlobot)...\n");

    /* Robot detach. */
    if((pid = fork()) > 0) {
      if(!net_init()) {
	perror("net_init()");
	kill(pid, SIGTERM);
	return 1;
      }
      if(!net_accept()) {
	perror("net_accept()");
	return 1;
      }
    } else {

      close(1);
      close(2);

      usleep(ROB_CONNECTDELAY);

      if(!net_connect(LOCALHOST))
	return 1;

      /* Hehe */
      argv[0] = "zlobot"; argv[1][0] = 0;

      /* Initialize the robot. */
      rob_init(difficulty);

      /* Robot's pace-negotiation routine. */
      rob_negotiate();

        /* Initialize alarm handler. */
      (void) signal(SIGALRM, robalarmhandle);

      /* Set the pace of the game. */
      ualarm(fps, 1000000 / fps);

      /* Initialize game parameters. */
      initgame();

      while(1) sleep(1);
    }
    
  } else {
    fprintf(stderr, "Connecting to %s...", argv[1]);
    while(!net_connect(argv[1])) {
      sleep(RECONNECT_DELAY);
      putc('.', stderr);
    }
  }

  /* Initialize screen. */
  if(!gui_init()) {
    fprintf(stderr, "Couldn't initialize display!\n");
    net_close();
    return 1;
  } 

  /* Display splash screen. */
  gui_splash_on();

  /* Initialize randomizer. */
  srandom(getpid() + getuid() + time(NULL));

  /* Initialize colors and so on. */
  if(!gfx_init()) {
    fprintf(stderr,
	    "Couldn't initialize graphics stuff. Probably out of mem.\n");
    net_close();
    gui_kill();
    return 1;
  }

  /* Generate background image. */ 
  mangle(); 

  /* Show off the splash for a bit. */
  if(!net || rob)
    sleep(SPLASH_DELAY);

  /* Negotiate terms with our partner. */
  if((fps = negotiate(net)) == -1) {
    fprintf(stderr, "Ident error with peer!\n");
    net_close();
    gui_kill();
    return 1;
  } else if(fps == -2) {
    fprintf(stderr, "Incompatible versions! Your friend should upgrade.\n");
    net_close();
    gui_kill();
    return 1;
  } else
    fprintf(stderr, "Calibrated game rate to %d ticks/sec.\n", fps);

  /* Initialize alarm handler. */
  (void) signal(SIGALRM, alarmhandle);

  /* Set the pace of the game. */
  ualarm(fps, 1000000 / fps);

  /* Initialize game parameters. */
  initgame();

  /* Switch display to game mode. */
  gui_splash_off();

  /* Let the game do its magic. */
  /* gui_event_loop();*/
  while(1) sleep(31337);

  return 0;
}



