/*
 * net.c
 * Network routines.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>

#include "config.h"
#include "proto.h"


static int ws, as, connected = 0;
static struct sockaddr_in wsin;



/* Send a string to a socket. */
int net_put(char *buf)
{
  if(!connected)
    return 1;

  if(send(as, buf, strlen(buf), 0) < 0) {
    net_close();
    return 0;
  }
  else {
    return 1;
  }
}


/* Read a string from a socket. */
int net_get(char *buf)
{
  int i;
	
  if(!connected)
    return 1;

  for(i = 0; i < BUFLEN; i++) {
    if(recv(as, buf + i, 1, 0) < 1) {
      net_close();
      return 0;
    }
    /* Strip carriage returns. */
    if(buf[i] == '\r') i--;
    
    /* We're done when we get a line feed. */
    else if(buf[i] == '\n') break;
  }
  
  /* Terminate the string. */
  buf[i] = 0;
  return 1;
}


/* Close a socket. */
int net_close()
{
  if(connected) {
    close(as);
#ifdef VERBOSE
    fprintf(stderr, "Disconnected.\n");
#endif
    connected = 0;

#ifdef QUIT_ON_DISC
    clean_exit();
#endif
  }

  return 0;
}


/* Connect to the given host. Returns whether it worked. */
int net_connect(char *host)
{
  struct sockaddr_in sin;
  struct hostent *hent;

  if((as = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return 0;

  memset(&sin, 0, sizeof(sin));

  if((hent = gethostbyname(host)) != NULL)
    memcpy(&sin.sin_addr.s_addr, hent->h_addr, hent->h_length);
  else if((sin.sin_addr.s_addr = inet_addr(host)) < 0) {
    close(as);
    return 0;
  }

  sin.sin_family = AF_INET;
  sin.sin_port = htons(PORT);
  if(connect(as, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
    close(as);
    return 0;
  }

#ifdef VERBOSE
  fprintf(stderr, " connected.\n");
#endif

  connected = 1;

  return 1;
}


/* Initialize the server socket. */
int net_init()
{
  int opt = 1;
				
  /* Allocate a socket. */
  if((ws = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    return 0;
  
  setsockopt(ws, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt));
  
  /* Fill in the socket structure. */
  memset(&wsin, 0, sizeof(wsin));
  wsin.sin_family = PF_INET;
  wsin.sin_port = htons(PORT);
  wsin.sin_addr.s_addr = htonl(inet_addr(LOCALHOST));
  
  /* Bind the socket to the port. */
  if(bind(ws, (struct sockaddr *)&wsin, sizeof(wsin)) < 0)
    return 0;
  
  /* Listen for a connection. */
  if(listen(ws, MAXQUEUE) < 0)
    return 0;
  
#ifdef VERBOSE
  fprintf(stderr, "Waiting for connection... ");
#endif

  return 1;
}


/* The main event loop. */
int net_snag()
{
  fd_set fdset;
  struct timeval tv;
  char buf[BUFLEN];
 
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  FD_ZERO(&fdset);
  FD_SET(as, &fdset);
  
  /* Wait for a change in the state of any descriptors. */
  while(select(as + 1, &fdset, NULL, NULL, &tv) > 0) {

    if(FD_ISSET(as, &fdset) && net_get(buf) < 0)
      return -1;

    update_us(buf);
  }

  return 0;
}


/* Accept the connection. */
int net_accept()
{
  struct sockaddr_in asin;
  int fromlen = sizeof(asin);
  
  /* Accept the connection. */
  if((as = accept(ws, (struct sockaddr *)&asin, &fromlen)) < 0)
    return 0; 
  
#ifdef VERBOSE
  fprintf(stderr, "connected to %s\n", inet_ntoa(asin.sin_addr));
#endif

  close(ws);

  connected = 1;

  return 1;
}

